import requests, json
import os.path
import re
import urllib.parse
import csv
from typing import Dict, Tuple
import pandas as pd

import regex_patterns

# Constants for the authentication and query process
AUTH_URL = 'https://app.leak.center/uaa/oauth/token'
LEAKS_ENDPOINT = "https://app.leak.center/svc-saas/leak/search"
CLIENT_ID = 'client-api'
CLIENT_SECRET = 'comfy-litigate-embargo-forelimb'
IS_DEBUG_REGEX = False


class Auth:
    username = None
    password = None


class ProcessLeaks:
    """
    """
    def __init__(self, domain: str, token: str, date_range: str, tags: list):
        self.domain = domain
        self.token = token
        self.date_range = date_range
        self.tags = tags
        self.is_debug = False
        self.is_debug_regex = IS_DEBUG_REGEX
        self._output_folder = 'output_files'
        self._empty_string = 'N/A'
        self._file = "{}/output_data_{}.csv".format(self._output_folder, self.domain)
        self._file_debug = "debug_files/output_data_{}.json".format(self.domain)
        self._init()

    def _init(self):
        if not os.path.isdir(self._output_folder):
            os.makedirs(self._output_folder)

        if not os.path.isdir('debug_files'):
            os.makedirs('debug_files')

    def run(self):
        if self.is_debug_regex:
            self._debug_regex()
            return

        count_all_results = self._run()
        self._remove_duplicates()
        return count_all_results

    def _run(self):
        debug_results = []
        page = 0
        count_all_results = 0
        with open(self._file, mode='w') as outfile:
            file_writer = csv.writer(outfile, delimiter=';', quotechar='"', quoting=csv.QUOTE_MINIMAL)
            file_writer.writerow(('URL', 'User', 'Password', 'leakPublishDate', 'fileName'))

            while True:
                params = self._get_query_params(page)
                data = self._send_request(params)
                filtered_content = filter_results(data['content'], self.tags, self.domain)

                if self.is_debug:
                    debug_results.extend(data['content'])

                for key, item in enumerate(filtered_content):
                    row = self._get_row(item)
                    file_writer.writerow(row)
                    # if key > 5:
                    #     break

                count_all_results += len(filtered_content)
                print(f"Page {page + 1} processed, total entries collected: {count_all_results}")  # Debugging output

                if data['last']:
                    break

                if self.is_debug:
                    if page >= 0:
                        break

                page += 1

        if self.is_debug:
            # print(json.dumps(debug_results, indent=4, sort_keys=True))
            with open(self._file_debug, 'w') as f:
                json.dump(debug_results, f, indent=4, sort_keys=True)

        return count_all_results

    def _get_query_params(self, page):
        params = {
            "page": page,
            "size": 100,
            "sort": "createdAt,desc",
            "length": 500,
            "highlight": "true",
            "query": self.domain + (" AND createdAt:[" + self.date_range + "]" if self.date_range else "")
        }
        return params

    def _remove_duplicates(self):
        """
        Remove duplicates in csv file base on columns in "subset" defined
        :return:
        """
        print("Start removing duplicates...")
        df = pd.read_csv(self._file, delimiter=';', keep_default_na=False) # .to_dict(orient='records')
        df.drop_duplicates(keep='first', inplace=True, subset=['URL', 'User', 'Password'])
        print("Updating the file...")
        df.to_csv(self._file, sep=";", index=False)
        print("Duplicates has been removed")

    def _send_request(self, params):
        # Correctly construct the full query URL with consistent string delimiters
        query_params = '&'.join(f"{k}={urllib.parse.quote(str(v), safe='')}" for k, v in params.items())
        query_url = f"{LEAKS_ENDPOINT}?{query_params}"
        print(f"Query URL: {query_url}")  # Debugging output: Show the actual query URL
        headers = {"Authorization": f"Bearer {self.token}"}
        response = requests.get(LEAKS_ENDPOINT, params=params, headers=headers)
        response.raise_for_status()
        data = response.json()
        return data

    def _get_row(self, item: Dict) -> tuple[str, str, str, str, str]:
        """
        Formatting csv row
        """
        url_title = self._empty_string
        user_title = ''
        pass_title = ''
        pub_date_title = ''
        file_name = ''

        for key, value in item.items():
            if key == 'Server':
                url_title = value if value else self._empty_string
            if key == 'User':
                user_title = self.clear_email(value)
            if key == 'Password':
                pass_title = value
            if key == 'createdAt':
                pub_date_title = value
            if key == 'fileName':
                file_name = value
        row = (url_title, user_title, pass_title, pub_date_title, file_name)
        return row

    def _debug_regex(self):
        data = {}
        data['content'] = [
            {
                "content": "ia.speedantre/;smkn1tab10otkp3;qwerty\nhttps://accounts.google.com/;kalyanisjunghari@gmail.com;Kalyani@2003\nhttps://sukienbts-freefire.vn/;PhuongNguyendnv@gmail.com;[NOT_SAVED]\nhttps://www.facebook.com;Trc Dung;EmGhetAnh80\nhttps://<kaduu:h>guest.buhlergroup.com</kaduu:h>/;gustavo.saito@buhlergroup.com;288341\nhttps://cls6.edunet.net/cyber/cm/mcom/pmco000b00.do;sje20346;sje20346!\nhttps://discord.com/login;arthurjmj274@gmail.com;97426394\nhttp://192.168.178.1/;;hold8605\nhttps://idp.omlet.me/register;ProPlayer1234;wr@Xbm",
                "createdAt": "2025-02-04 13:06:46",
                "cvssScore": 7.0,
                "fileExtension": "txt",
                "fileName": "stelear logs 01.02.2025/2025.02.012318freelogpassforallcloudsbase251.txt",
                "id": "1785566a-d0aa-3352-b6d1-ae151e885c7a",
                "leakDiscoverDate": "2025-02-03",
                "leakId": "ca0856e8-50be-398a-8119-612e3be65a11",
                "leakName": "stelear logs 01.02.2025",
                "leakPublishDate": "2025-02-01",
                "leakSize": 5408049582,
                "leakSource": "darknet",
                "leakTags": "password,log,email,hash,url,account,username"
            }
        ]
        filtered_content = filter_results(data['content'], self.tags, self.domain)
        for key, item in enumerate(filtered_content):
            row = self._get_row(item)
            print('--row--')
            print(row)

    def clear_email(self, email):
        match = re.search(regex_patterns.get_email_group_pattern(), email, re.IGNORECASE)
        if match:
            result = match.groupdict()
            email = result.get('Email')

        return email


def get_credentials():
    if Auth.username and Auth.password:
        return Auth.username, Auth.password
    # if not os.path.isfile("auth.txt"):
    #     raise ValueError("auth.txt not found")
    try:
        with open("auth.txt", "r") as f:
            return f.read().strip().split(',')
    except FileNotFoundError:
        raise ValueError("auth.txt not found")


def get_access_token(username, password):
    data = {
        'grant_type': 'password',
        'username': username,
        'password': password,
        'client_id': CLIENT_ID,
        'client_secret': CLIENT_SECRET
    }
    response = requests.post(AUTH_URL, data=data)
    response.raise_for_status()
    token = response.json()['access_token']
    print(f"Success Authentication, Fetched token: {token[:10]}...")  # Show only the first 10 characters for security
    return token


def get_query_input():
    try:
        with open("input.txt", "r") as f:
            return f.read().strip()
    except FileNotFoundError:
        return  # This should be replaced with actual in-memory variable if available


def get_date_range():
    try:
        with open("DateRange.txt", "r") as f:
            return f.read().strip()
    except FileNotFoundError:
        return ""  # This should be replaced with actual in-memory variable if available


def get_tags():
    with open("tags.txt", "r") as f:
        return [line.strip() for line in f]


def query_and_process_leaks(domain, token, date_range, tags):
    """
    """
    process_leaks = ProcessLeaks(domain, token, date_range, tags)
    count_results = process_leaks.run()
    return count_results


def is_valid_server(server):
    """
    Check if the server string is a valid URL format.
    """
    valid_server_pattern = re.compile(
        r"^(https?://|http://|android://|ftp://|ttps://|ttp://|tps://|tp://|ps://|s://|p://|oid://)?([a-zA-Z0-9_-]+(\.[a-zA-Z0-9_-]+)+)(:[0-9]+)?(/[^\s]*)?$")
    return bool(valid_server_pattern.match(server))


def extract_info(input_string, domain):
    for pattern in regex_patterns.get_patterns(domain):
        match = re.search(pattern, input_string)
        if match:
            if IS_DEBUG_REGEX:
                print('---Domain---', domain)
                print('---Input String---', input_string)
                print('---Pattern---', pattern)
            result = match.groupdict()
            if result.get('Server') and not is_valid_server(result['Server']):
                result['Server'] = None
            if result.get('Server'):
                result['Server'] = re.sub(r'^(ttps|ttp|tps|tp|ps|s|p|oid)://', '', result['Server'])
            if result.get('User') and 'https://' in result.get('User'):
                result['User'] = None

            extracted_info = dict()
            for key in ['Server', 'User', 'Password']:
                value = result.get(key, "")
                if value is None or value.lower() in ["none", ":", ";", ",", "null", "empty", "unknown"]:
                    extracted_info[key] = ""
                else:
                    extracted_info[key] = value

            return extracted_info

    return {'Server': "", 'User': "", 'Password': ""}


def filter_results(results, tags, domain):
    filtered_content = []
    tag_pattern = re.compile("|".join(tags))
    for result in results:
        try:
            tag_found = tag_pattern.search(result['leakTags'])
        except Exception:
            tag_found = False
        if tag_found:
            content_lines = result['content'].split('\n')
            for line in content_lines:
                if "<kaduu:h>" in line:
                    line = cut_sql_content(line)

                    clean_line = re.sub(r'<kaduu:h>|</kaduu:h>|\\|^\s', '', line)  # fix1 |^\s
                    data = {
                        "LeakContent": clean_line,
                        "createdAt": result['createdAt'],
                        "fileName": result['fileName'],
                        # "leakSource": result['leakSource'],
                        # "leakSize": result['leakSize'],
                        # "leakPublishDate": result['leakPublishDate'],
                        # "leakTags": result['leakTags']
                    }
                    info = extract_info(clean_line, domain)

                    for key, value in info.items():
                        if value:
                            data[key] = value
                    filtered_content.append(data)
        else:
            print(f"No matching tags found in leakTags: {result['leakTags']}")  # Debugging output for no match
    return filtered_content


def cut_sql_content(line):  # todo to regex
    match = re.search("\),\(", line)
    if match:
        sql_lines = line.split("),(")
        for sql_line in sql_lines:
            if "<kaduu:h>" in sql_line:
                line = sql_line

    match = re.search("\\\\r\\\\", line)
    if match:
        sql_lines = line.split("\\r\\")
        for sql_line in sql_lines:
            if "<kaduu:h>" in sql_line:
                line = sql_line
    return line


def main():
    token = None
    if IS_DEBUG_REGEX is False:
        try:
            username, password = get_credentials()
        except ValueError as e:
            username = input('No auth.txt found or login invalid. Please enter your username (email):')
            password = input('Enter your password:')

        if not username or not password:
            print('Username and password are required values')
            return

        try:
            token = get_access_token(username, password)
        except Exception as e:
            print('Can not get access token. Authentication is failed. Check you username/password')
            return

    domain = get_query_input()

    if not domain:
        domain = input('Enter your domain like example.com (no subdomain):')

    if not domain:
        print('Domain is empty')
        return

    date_range = get_date_range()

    if not date_range:
        # format 2024-01-01 TO *
        date_range = input('Enter your date range in the syntax of 2024-02-01 TO * or 2024-02-01 TO 2025-01-31 or press enter to ignore date range:')

    tags = get_tags()
    final_results = query_and_process_leaks(domain, token, date_range, tags)

    print("Completed. Results saved to output_files/output_data_{REQUESTED_DOMAIN}.csv")


if __name__ == "__main__":
    main()
